//<vertex>
	void main()
	{
		gl_Position = ftransform();
		gl_TexCoord[0] = gl_MultiTexCoord0;
	}
//</vertex>

//<fragment>
	uniform sampler2D samFront;
	uniform sampler2D samBack;
	uniform sampler3D samVolume;
	uniform sampler2D samTransferFunction;
	uniform sampler2D samNoise;
	uniform sampler3D samAmbientTexture;

	uniform int ray_steps;
	uniform int n_rays;
	uniform int steplength;
	uniform float z_cut;
	uniform float y_cut;
	uniform float x_cut;
	uniform float fSampleDistance;
	uniform float fUnitStep;
	uniform vec3 vecSteps;
	uniform vec3 vecLight;
	uniform vec3 vecHalf;
	uniform mat4 matVolume;
	uniform mat4 matInverse;
  
	vec3 vecRayDirection;
  
	//<option name="components" value="1">
	
		vec4 GetColor(vec3 vecPosition)
		{
			vec4 vecSample;

			vec3 vecSampleOne;
			vecSampleOne.x = texture3D(samVolume,vecPosition-vec3(vecSteps.x,0.0,0.0)).a;
			vecSampleOne.y = texture3D(samVolume,vecPosition-vec3(0.0,vecSteps.y,0.0)).a;
			vecSampleOne.z = texture3D(samVolume,vecPosition-vec3(0.0,0.0,vecSteps.z)).a;
			vecSampleOne *= 0.5;

			vec3 vecSampleTwo;
			vecSampleTwo.x = texture3D(samVolume,vecPosition+vec3(vecSteps.x,0.0,0.0)).a;
			vecSampleTwo.y = texture3D(samVolume,vecPosition+vec3(0.0,vecSteps.y,0.0)).a;
			vecSampleTwo.z = texture3D(samVolume,vecPosition+vec3(0.0,0.0,vecSteps.z)).a;
			vecSampleTwo *= 0.5;

			vecSample.xyz = vecSampleOne - vecSampleTwo;
			vecSample.a = 0.25 * (texture3D(samVolume,vecPosition).a + vecSampleOne.x + vecSampleOne.y + vecSampleOne.z + vecSampleTwo.x + vecSampleTwo.y + vecSampleTwo.z);
						
			vec4 vecColor;
			vecColor = texture2D(samTransferFunction,vec2(vecSample.a,0.5));
	

			if (vecColor.a > 0.0)
			{   
				vec3 vecNormal = (matVolume * vec4(vecSample.x,vecSample.y,vecSample.z,0.0)).xyz;
				float fLength = length(vecNormal);

				if (fLength > 0.0)
				{
					vecNormal /= fLength;
					float fDiffuse = abs(dot(vecLight,vecNormal));
					vecColor.rgb *= fDiffuse;

					float fSpecular = pow(abs(dot(vecHalf,vecNormal)),32.0);
					vecColor.rgb += fSpecular;
				}
			}
			//vecColor = vec4(0.0, 1.0, 0.0, 1.0);
			return vecColor;
		}
		
		
		
		
		
		vec4 GetAmbient(vec3 vecPosition)
		{
			vec4 vecSample;

			vec3 vecSampleOne;
			vecSampleOne.x = texture3D(samVolume,vecPosition-vec3(vecSteps.x,0.0,0.0)).a;
			vecSampleOne.y = texture3D(samVolume,vecPosition-vec3(0.0,vecSteps.y,0.0)).a;
			vecSampleOne.z = texture3D(samVolume,vecPosition-vec3(0.0,0.0,vecSteps.z)).a;
			vecSampleOne *= 0.5;

			vec3 vecSampleTwo;
			vecSampleTwo.x = texture3D(samVolume,vecPosition+vec3(vecSteps.x,0.0,0.0)).a;
			vecSampleTwo.y = texture3D(samVolume,vecPosition+vec3(0.0,vecSteps.y,0.0)).a;
			vecSampleTwo.z = texture3D(samVolume,vecPosition+vec3(0.0,0.0,vecSteps.z)).a;
			vecSampleTwo *= 0.5;

			vecSample.xyz = vecSampleOne - vecSampleTwo;
			vecSample.a = 0.25 * (texture3D(samVolume,vecPosition).a + vecSampleOne.x + vecSampleOne.y + vecSampleOne.z + vecSampleTwo.x + vecSampleTwo.y + vecSampleTwo.z);
						
			vec4 vecColor;
			vecColor = texture2D(samTransferFunction,vec2(vecSample.a,0.5));
	
			return vecColor;
		} 
		
		
	//</option>

	//<option name="components" value="3">
		vec4 GetColor(vec3 vecPosition)
		{
			vec4 vecSample;

			vecSample.rgb = texture3D(samVolume,vecPosition).rgb;

			vec3 vecX = 0.5 * (texture3D(samVolume,vecPosition-vec3(vecSteps.x,0.0,0.0)).rgb - texture3D(samVolume,vecPosition+vec3(vecSteps.x,0.0,0.0)).rgb);
			vec3 vecY = 0.5 * (texture3D(samVolume,vecPosition-vec3(0.0,vecSteps.y,0.0)).rgb - texture3D(samVolume,vecPosition+vec3(0.0,vecSteps.y,0.0)).rgb);
			vec3 vecZ = 0.5 * (texture3D(samVolume,vecPosition-vec3(0.0,0.0,vecSteps.z)).rgb - texture3D(samVolume,vecPosition+vec3(0.0,0.0,vecSteps.z)).rgb);

			vec3 vecGradient;
			vecGradient.x = 0.33 * (vecX.r + vecX.g + vecX.b);
			vecGradient.y = 0.33 * (vecY.r + vecY.g + vecY.b);
			vecGradient.z = 0.33 * (vecZ.r + vecZ.g + vecZ.b);

			vecSample.a = length(vecGradient);
			return vecSample;
		}
	//</option>

	//<option name="components">
		vec4 GetColor(vec3 vecPosition)
		{
			return vec4(0.0,0.0,0.0,0.0);
		}
	//</option>

	void main()
	{
		vec4 vecRayColor = vec4(0.0,0.0,0.0,0.0);
		vec4 vecFront = texture2D(samFront,gl_TexCoord[0].xy);
		vec4 vecBack = texture2D(samBack,gl_TexCoord[0].xy);
				
		gl_FragDepth = 1.0;

		if (vecBack.a > vecFront.a)
		{
			float fDistance = distance(vecFront.xyz,vecBack.xyz);
			vec4 vecRayIncrement = vec4((vecBack.xyz - vecFront.xyz) / fDistance, 1.0) * (fSampleDistance*fUnitStep);

			vecRayDirection = normalize(vecBack.xyz - vecFront.xyz);			
			vec4 vecRayPosition = vec4(vecFront.xyz,0.0);

			float fNoise = 2.0 * texture2D(samNoise,gl_FragCoord.xy / 64.0).a - 1.0;
			vecRayPosition += fNoise*vecRayIncrement;
			
			bool bContinue = false;

			//<option name="technique">
			do
			{   
				vec4 vecSampleColor = GetColor(vecRayPosition.xyz);		

				vecRayPosition += vecRayIncrement;
				bContinue = vecRayPosition.w < fDistance;

				if (vecSampleColor.a > 0.0 && vecRayPosition.z>z_cut && vecRayPosition.y>y_cut && vecRayPosition.x>x_cut)
					break;

			}
			while (bContinue);

			if (bContinue)
			{
				vec4 vecFirstHit = matInverse*vec4(vecRayPosition.xyz,1.0);
				vecFirstHit.xyz /= vecFirstHit.w;
				gl_FragDepth = (vecFirstHit.z+1.0) / 2.0;

				do
				{
					do
					{
						vec4 vecSampleColor = GetColor(vecRayPosition.xyz);		

						if (vecSampleColor.a > 0.0)
						{
							vecSampleColor.a = 1.0 - pow(1.0-vecSampleColor.a,fSampleDistance);
							vecSampleColor.a *= (1.0 - vecRayColor.a);
							vecSampleColor.rgb *= vecSampleColor.a;
							vecRayColor += vecSampleColor;
						}

						vecRayPosition += vecRayIncrement;
						bContinue = vecRayPosition.w < fDistance && vecRayColor.a < 1.0;
					}
					while (bContinue);
				}
				while (bContinue);
			}

			//</option>

			//<option name="technique" value="MIP">
			float fMaximumValue = -1.0;
			vec4 vecMaximumPosition = vec4(0.0,0.0,0.0,0.0);

			do
			{
				do
				{
					float fValue = texture3D(samVolume,vecRayPosition.xyz).a;

					if (fValue > fMaximumValue)
					{
						fMaximumValue = fValue;
						vecMaximumPosition = vecRayPosition;
					}

					vecRayPosition += vecRayIncrement;
					bContinue = vecRayPosition.w < fDistance;
				}
				while (bContinue);
			}
			while (bContinue);

			vecRayColor = texture2D(samTransferFunction,vec2(fMaximumValue,0.5));

			if (vecRayColor.a > 0.0)
			{
				vec4 vecFirstHit = matInverse*vec4(vecMaximumPosition.xyz,1.0);
				vecFirstHit.xyz /= vecFirstHit.w;
				gl_FragDepth = (vecFirstHit.z+1.0) / 2.0;
			}

			//</option>
			
			//###############################################################################
			// ambient starts here
			//###############################################################################

			//<option name="technique" value="Ambient">

			do
			{   
				vec4 vecSampleColor = GetAmbient(vecRayPosition.xyz);		

				vecRayPosition += vecRayIncrement;
				bContinue = vecRayPosition.w < fDistance;

				if (vecSampleColor.a > 0.0 && vecRayPosition.z>z_cut && vecRayPosition.y>y_cut && vecRayPosition.x>x_cut)
					break;

			}
			while (bContinue);

			if (bContinue)
			{
				vec4 vecFirstHit = matInverse*vec4(vecRayPosition.xyz,1.0);
				vecFirstHit.xyz /= vecFirstHit.w;
				gl_FragDepth = (vecFirstHit.z+1.0) / 2.0;

				do
				{
					do
					{
						vec4 vecSampleColor = GetAmbient(vecRayPosition.xyz);		


            if (vecSampleColor.a > 0.0){    //point is visible, need to calculate shadow then
              //vecSampleColor.x=vecRayPosition.x;
              //vecSampleColor.y=vecRayPosition.y;
              //vecSampleColor.z=vecRayPosition.z;
              


              vec4 tempPoint;
              vec4 samplingRays[26];
              float vecSize = 1.0/128.0;
              float vecSizeN = 1.0/(128.0*pow(3.0, 0.5)); //shorter because of diagonal
              
              
              float occlusion = 0.0;
	        
              //along axis    
              samplingRays[0]=vec4(vecSize, 0.0, 0.0, 0.0); 
              samplingRays[1]=vec4(0.0, vecSize, 0.0, 0.0); 
              samplingRays[2]=vec4(0.0, 0.0, vecSize, 0.0); 
              samplingRays[3]=vec4(-vecSize, 0.0, 0.0, 0.0); 
              samplingRays[4]=vec4(0.0, -vecSize, 0.0, 0.0); 
              samplingRays[5]=vec4(0.0, 0.0, -vecSize, 0.0);               
              
              //45 degree from axis, not normalized
              samplingRays[6]=vec4(vecSizeN, vecSizeN, vecSizeN, 0.0); 
              samplingRays[7]=vec4(vecSizeN, vecSizeN, -vecSizeN, 0.0); 
              samplingRays[8]=vec4(vecSizeN, -vecSizeN, vecSizeN, 0.0); 
              samplingRays[9]=vec4(vecSizeN, -vecSizeN, -vecSizeN, 0.0); 
              samplingRays[10]=vec4(-vecSizeN, vecSizeN, vecSizeN, 0.0); 
              samplingRays[11]=vec4(-vecSizeN, vecSizeN, -vecSizeN, 0.0); 
              samplingRays[12]=vec4(-vecSizeN, -vecSizeN, vecSizeN, 0.0); 
              samplingRays[13]=vec4(-vecSizeN, -vecSizeN, -vecSizeN, 0.0); 
              
              //middle of an edge of a cube, not normalized for now
              samplingRays[14]=vec4(-vecSize, -vecSize, 0.0, 0.0); 
              samplingRays[15]=vec4(-vecSize, 0.0, -vecSize, 0.0); 
              samplingRays[16]=vec4(-vecSize, 0.0, vecSize, 0.0); 
              samplingRays[17]=vec4(-vecSize, vecSize, 0.0, 0.0); 
              
              samplingRays[18]=vec4(0.0, -vecSize, 0.0, 0.0); 
              samplingRays[19]=vec4(0.0, 0.0, -vecSize, 0.0); 
              samplingRays[20]=vec4(0.0, 0.0, vecSize, 0.0); 
              samplingRays[21]=vec4(0.0, vecSize, 0.0, 0.0); 

              samplingRays[22]=vec4(vecSize, -vecSize, 0.0, 0.0); 
              samplingRays[23]=vec4(vecSize, 0.0, -vecSize, 0.0); 
              samplingRays[24]=vec4(vecSize, 0.0, vecSize, 0.0); 
              samplingRays[25]=vec4(vecSize, vecSize, 0.0, 0.0); 
              
             /* 
              #########################
              
               //simple, just checking 14 points, random?
              int n_directions=26;  
              float n_directions_f=26.0;  
              const int n_steps = 20; //need const to create array later
              const float n_steps_f = 20.0;
              
              float sumvalue = 0.0;

              
              #####################
              */

              //<option name="selectable" value="NoSelect14">
					//simple, just checking 14 points, random?
					int n_directions=14;  
					float n_directions_f=14.0;  
					const int step_max = 20;
					int n_steps = 20;
					float n_steps_f=20;
              //</option>


              //<option name="selectable" value="NoSelect26">
					//simple, just checking 14 points, random?
					int n_directions=26;  
					float n_directions_f=26.0;  
					const int step_max = 20;
					int n_steps = 20;
					float n_steps_f=20;
              //</option>

              
              //<option name="selectable" value="Select">
              
					int n_directions=n_rays;
					float n_directions_f=0.0;
					for(int i =0; i<=n_directions;i++){
						n_directions_f=n_directions_f+1.0;
					} 
					const int step_max = 40; //length of the array, 100 not possible to use, because of time
					//const int n_steps = 10; //need const to create array later
					int n_steps = ray_steps;
		            float n_steps_f=0.0;
					for(int i =0; i<=n_steps;i++){
						n_steps_f=n_steps_f+1.0;
					} 
		            //const float n_steps_f = 10.0;
      		  //</option>
              
              float sumvalue = 0.0;
              
              for (int i=0;i<n_directions;i++){
//------------------------------------                
              //int i = 1; //just one direction for now
              
              float distCounter=1.0;
              float rayvalues[step_max];
              float raymulti[step_max]; //to reduce multiplications
              for(int j=0;j<n_steps;j++){ //placed stuff in table, should not include startpoint?
                rayvalues[j]= (GetAmbient(vecRayPosition.xyz+distCounter*samplingRays[i].xyz)).a; //store .a value
                raymulti[j] = 0.0; //to clear it
                distCounter=distCounter + 1.0;
                }
    
                
              //main loop, must be shortened
              for(int j=0;j<n_steps;j++){
                float temp =1.0;
                for(int k=0;k<j;k++){  //from i = 0 to m-1
                  temp=temp*(1.0-rayvalues[k]);
                  //temp=temp*rayvalues[k];
                }
                raymulti[j]=temp;  
              }
                
    
/*                
              raymulti[0]=1.0; //init value needs to be one
              for(int j=1;j<n_steps;j++){
                raymulti[j]=raymulti[j-1]*(1.0-rayvalues[j-1]);
              }
  */            
              float tempvalue=0.0;
              //start from 1 because first place should be empty
              for(int j=1;j<n_steps;j++){
                tempvalue=tempvalue+raymulti[j];
              }
              tempvalue=tempvalue/n_steps_f;
              sumvalue=sumvalue+tempvalue;
//------------------------------------
              }
              sumvalue=sumvalue/n_directions_f;
              
              //float distCounter=1.0;
              //float n_directions_f=0.0;
              //for(int i =0; i<n_directions; i++){
                //n_directions_f=n_directions_f+1.0; //float counter
                //for(int j=0;j<30;j++){
                 // tempPoint=GetAmbient(vecRayPosition.xyz+distCounter*samplingRays[i].xyz);
          //        if(tempPoint.a>0.0){ //have occluding object here, stop adding?
          //          break;
          //        }
                  //occlusion=occlusion+tempPoint.a/distCounter;
                  //occlusion=occlusion+tempPoint.a;
                 // distCounter=distCounter+1.0;
                //}
                //distCounter=1.0;
             // }
             // occlusion=occlusion/(n_directions_f*5.0); //6 points
             // vecSampleColor.xyz=vecSampleColor.xyz*(1.0-occlusion);

             // vecSampleColor.xyz=vecRayPosition.xzx;
              
              vecSampleColor.xyz=vecSampleColor.xyz*sumvalue;
              
            }//end of: if (vecSampleColor.a > 0.0){
              
            
            
            
            

						if (vecSampleColor.a > 0.0)
						{
							vecSampleColor.a = 1.0 - pow(1.0-vecSampleColor.a,fSampleDistance);
							vecSampleColor.a *= (1.0 - vecRayColor.a);
							vecSampleColor.rgb *= vecSampleColor.a;
							vecRayColor += vecSampleColor;
						}

						vecRayPosition += vecRayIncrement;
						bContinue = vecRayPosition.w < fDistance && vecRayColor.a < 1.0;
					}
					while (bContinue);
				}
				while (bContinue);
			}





			
			//</option>
			//###############################################################################
			// ambient ends here
			//###############################################################################
			
		}
		else
			discard;

		vecRayColor.rgb *= vecRayColor.a;
		gl_FragColor = vecRayColor;
		
	}
//</fragment>
